#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import List, FrozenSet, Set, Optional
import re

from gambatools.automaton import Automaton


def default_state_label_regex() -> str:
    return r'\w+'


def default_transition_label_regex() -> str:
    return r'.+'


def default_symbol_regex() -> str:
    return r'\w+'


def state_set_regex() -> str:
    return r'\{[\w,]*\}'


def state_product_regex() -> str:
    return r'\(\w+,\w+\)'


def dfa_keywords() -> FrozenSet[str]:
    return frozenset(['input_symbols'])


def nfa_keywords() -> FrozenSet[str]:
    return frozenset(['input_symbols', 'epsilon'])


def pda_keywords() -> FrozenSet[str]:
    return frozenset(['input_symbols', 'stack_symbols', 'epsilon'])


def tm_keywords() -> FrozenSet[str]:
    return frozenset(['input_symbols', 'tape_symbols', 'blank', 'accept', 'reject'])


def automaton_keywords() -> FrozenSet[str]:
    return dfa_keywords() | nfa_keywords() | pda_keywords() | tm_keywords()


class AutomatonBuilder(object):
    def __init__(self, A: Automaton, state_regex = default_state_label_regex(), transition_regex=default_transition_label_regex(), symbol_regex=default_symbol_regex()):
        self.A = A
        self.state_regex = state_regex
        self.transition_regex = transition_regex
        self.symbol_regex = symbol_regex

    def _fresh_state(self, states: Set[str], hint: str = 'P') -> str:
        if hint not in states:
            return hint
        index = 1
        while True:
            state = '{}{}'.format(hint, index)
            if state not in states:
                return state
            index = index + 1

    def _check_symbol(self, symbol):
        if not re.fullmatch(self.symbol_regex, symbol):
            raise RuntimeError('invalid symbol {}'.format(symbol))

    def _check_symbols(self, symbols):
        for symbol in symbols:
            self._check_symbol(symbol)

    def _check_state_label(self, state) -> None:
        if not re.fullmatch(self.state_regex, state):
            raise RuntimeError('invalid state label {}'.format(state))

    def _check_transition_label(self, label) -> None:
        if not re.fullmatch(self.transition_regex, label):
            raise RuntimeError('invalid transition label {}'.format(label))

    def _check_state_labels(self):
        A = self.A
        for state in A.states:
            self._check_state_label(state)

    def _check_one_initial_state(self):
        A = self.A
        if len(A.initial_states) == 0:
            raise RuntimeError('the automaton has no initial state')
        elif len(A.initial_states) > 1:
            raise RuntimeError('the automaton has multiple initial states')

    def _check_states_are_declared(self):
        A = self.A
        used_states = A.used_states()
        if len(used_states - A.states) > 0:
            raise RuntimeError('the following states are not declared: {}'.format(', '.join(used_states - A.states)))

    def _check_symbols_are_declared(self, used_symbols: Set[str], declared_symbols: Set[str]):
        if len(used_symbols - declared_symbols) > 0:
            raise RuntimeError('the following symbols are not declared: {}'.format(', '.join(used_symbols - declared_symbols)))

    def get_symbol(self, key: str, default_value: str) -> str:
        A = self.A
        if key in A.items:
            if len(A.items[key]) == 0:
                raise RuntimeError('no value was specified for "{}"'.format(key))
            if len(A.items[key]) > 1:
                print('A.items =', A.items)
                raise RuntimeError('multiple values were specified for "{}"'.format(key))
            return A.items[key][0]
        return default_value

    def parse_symbol(self, key='epsilon', value='ε', default_value='_'):
        A = self.A
        if key in A.items:
            return self.get_symbol(key, default_value)
        for (p, a, q) in A.transitions:
            if value in a:
                return value
        return default_value


    def get_symbol_set(self, key: str, used_symbols: Optional[Set[str]] = None) -> Set[str]:
        A = self.A
        if key in A.items:
            declared_symbols = set(A.items[key])
            if used_symbols:
                self._check_symbols_are_declared(used_symbols, declared_symbols)
            return declared_symbols
        return used_symbols

    def get_state(self, key: str, default_value: str) -> str:
        A = self.A
        if key in A.items:
            if len(A.items[key]) == 0:
                raise RuntimeError('no value was specified for "{}"'.format(key))
            if len(A.items[key]) > 1:
                raise RuntimeError('multiple values were specified for "{}"'.format(key))
            return A.items[key][0]
        return default_value

    def build(self) -> Automaton:
        A = self.A
        A.states = A.states if len(A.states) > 0 else A.used_states()
        self._check_states_are_declared()
        self._check_state_labels()
        return A


class AutomatonParser(object):
    def __init__(self, transition_regex=default_transition_label_regex(), state_regex=default_state_label_regex(), keywords: FrozenSet[str] = automaton_keywords()):
        self.keywords = keywords
        self.state_regex = state_regex
        self.transition_regex = transition_regex
        self.items = {}
        self.states: Set[str] = set([])
        self.transitions = []
        self.initial_states: Set[str] = set([])
        self.final_states: Set[str] = set([])

    def _check_no_duplicate_keys(self, key: str):
        if key in self.items:
            raise RuntimeError('the keyword "{}" is specified multiple times'.format(key))

    def _check_no_duplicates(self, words: List[str]) -> None:
        W = set([])
        for word in words:
            if word in W:
                raise RuntimeError('duplicate entry "{}" found in "{}"'.format(word, words))
            W.add(word)

    def _check_keys_exist(self, keys: List[str]):
        for key in keys:
            if not key in self.items:
                raise RuntimeError('the keyword "{}" is missing'.format(key))

    def _check_state_label(self, state) -> None:
        if not re.fullmatch(self.state_regex, state):
            raise RuntimeError('invalid state label {}'.format(state))

    def _check_transition_label(self, label) -> None:
        if not re.fullmatch(self.transition_regex, label):
            raise RuntimeError('invalid transition label {}'.format(label))

    def parse_state(self, state: str) -> str:
        self._check_state_label(state)
        return state

    def parse_state_set(self, keyword, words: List[str], check_non_empty: bool = False) -> Set[str]:
        self._check_no_duplicate_keys(keyword)
        self._check_no_duplicates(words)
        if check_non_empty and len(words) == 0:
            raise RuntimeError('the set "{}" may not be empty'.format(keyword))
        return set(self.parse_state(word) for word in words)

    def parse_transition_label(self, label: str) -> str:
        self._check_transition_label(label)
        return label

    def parse_transition(self, words: List[str], line: str):
        if len(words) <= 2:
            raise RuntimeError('incomplete transition "{}"'.format(line))
        p = self.parse_state(words[0])
        q = self.parse_state(words[1])
        for a in words[2:]:
            a = self.parse_transition_label(a)
            self.transitions.append((p, a, q))

    def parse_line(self, line: str):
        words = line.strip().split()
        if not words or words[0].startswith('%'):
            pass
        elif words[0] == 'states':
            self.states = self.parse_state_set('states', words[1:], check_non_empty=True)
            self.items['states'] = words[1:]
        elif words[0] == 'final':
            self.final_states = self.parse_state_set('final', words[1:])
            self.items['final'] = words[1:]
        elif words[0] == 'initial':
            self.initial_states = self.parse_state_set('initial', words[1:])
            self.items['initial'] = words[1:]
        elif words[0] in self.keywords:
            keyword = words[0]
            self._check_no_duplicate_keys(keyword)
            self.items[keyword] = words[1:]
        else:
            self.parse_transition(words, line)

    def parse(self, text: str) -> Automaton:
        lines = text.split('\n')
        for line in lines:
            self.parse_line(line)
        return Automaton(self.states, self.transitions, self.initial_states, self.final_states, self.items)


def parse_automaton(text: str, transition_regex=default_transition_label_regex(), state_regex=default_state_label_regex(), symbol_regex=default_symbol_regex()) -> Automaton:
    A = AutomatonParser(state_regex=state_regex, transition_regex=transition_regex).parse(text)
    return AutomatonBuilder(A, state_regex=state_regex, transition_regex=transition_regex, symbol_regex=symbol_regex).build()



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Any, Set, List

import gambatools.dfa
import gambatools.dfa_algorithms
import gambatools.nfa
import gambatools.nfa_algorithms
import gambatools.pda
import gambatools.pda_algorithms
import gambatools.tm
import gambatools.tm_algorithms
import gambatools.cfg
import gambatools.cfg_algorithms
import gambatools.regexp
import gambatools.regexp_algorithms


def generate_language(L: Any, n: int) -> Set[str]:
    if isinstance(L, set) and all(isinstance(w, str) for w in L):
        result = L
    elif isinstance(L, gambatools.dfa.DFA):
        result = gambatools.dfa_algorithms.dfa_words_up_to_n(L, n)
    elif isinstance(L, gambatools.nfa.NFA):
        result = gambatools.nfa_algorithms.nfa_words_up_to_n(L, n)
    elif isinstance(L, gambatools.pda.PDA):
        result = gambatools.pda_algorithms.pda_words_up_to_n(L, n)
    elif isinstance(L, gambatools.tm.TM):
        result = gambatools.tm_algorithms.tm_words_up_to_n(L, n)
    elif isinstance(L, gambatools.cfg.CFG):
        result = gambatools.cfg_algorithms.cfg_words_up_to_n(L, n)
    elif isinstance(L, gambatools.regexp.Regexp):
        result = gambatools.regexp_algorithms.regexp_words_up_to_n(L, n)
    else:
        raise RuntimeError('cannot generate a language for type {}'.format(L.__class__))
    return result


# A1 is the user supplied answer
# A2 is the expected result
def compare_languages(A1: Set[str], A2: Set[str]) -> List[str]:
    A1minusA2 = sorted(A1 - A2, key=lambda x: (len(x)))
    A2minusA1 = sorted(A2 - A1, key=lambda x: (len(x)))
    feedback = []
    if len(A1minusA2) > 0:
        word = A1minusA2[0]
        word = 'ε' if not word else word
        feedback.append("Error: word '{}' should not be accepted".format(word))
    elif len(A2minusA1) > 0:
        word = A2minusA1[0]
        word = 'ε' if not word else word
        feedback.append("Error: word '{}' should be accepted".format(word))
    return feedback


# A1 is the user supplied answer
# A2 is the expected result
def check_equal_languages(L1: Any, L2: Any, length: int = 4) -> List[str]:
    A1 = generate_language(L1, length)
    A2 = generate_language(L2, length)
    return compare_languages(A1, A2)



# Generated from CFG.g4 by ANTLR 4.8
# encoding: utf-8
from antlr4 import *
from io import StringIO
import sys
if sys.version_info[1] > 5:
	from typing import TextIO
else:
	from typing.io import TextIO


def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\t")
        buf.write("9\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7\4\b")
        buf.write("\t\b\4\t\t\t\3\2\3\2\5\2\25\n\2\3\3\3\3\3\3\7\3\32\n\3")
        buf.write("\f\3\16\3\35\13\3\3\4\3\4\3\5\3\5\3\6\3\6\3\6\7\6&\n\6")
        buf.write("\f\6\16\6)\13\6\3\7\3\7\3\b\3\b\3\b\7\b\60\n\b\f\b\16")
        buf.write("\b\63\13\b\3\t\3\t\3\t\3\t\3\t\2\2\n\2\4\6\b\n\f\16\20")
        buf.write("\2\3\4\2\4\4\b\b\2\64\2\22\3\2\2\2\4\26\3\2\2\2\6\36\3")
        buf.write("\2\2\2\b \3\2\2\2\n\"\3\2\2\2\f*\3\2\2\2\16,\3\2\2\2\20")
        buf.write("\64\3\2\2\2\22\24\5\4\3\2\23\25\7\3\2\2\24\23\3\2\2\2")
        buf.write("\24\25\3\2\2\2\25\3\3\2\2\2\26\33\5\20\t\2\27\30\7\3\2")
        buf.write("\2\30\32\5\20\t\2\31\27\3\2\2\2\32\35\3\2\2\2\33\31\3")
        buf.write("\2\2\2\33\34\3\2\2\2\34\5\3\2\2\2\35\33\3\2\2\2\36\37")
        buf.write("\t\2\2\2\37\7\3\2\2\2 !\7\b\2\2!\t\3\2\2\2\"\'\5\6\4\2")
        buf.write("#$\7\5\2\2$&\5\6\4\2%#\3\2\2\2&)\3\2\2\2\'%\3\2\2\2\'")
        buf.write("(\3\2\2\2(\13\3\2\2\2)\'\3\2\2\2*+\5\n\6\2+\r\3\2\2\2")
        buf.write(",\61\5\f\7\2-.\7\6\2\2.\60\5\f\7\2/-\3\2\2\2\60\63\3\2")
        buf.write("\2\2\61/\3\2\2\2\61\62\3\2\2\2\62\17\3\2\2\2\63\61\3\2")
        buf.write("\2\2\64\65\5\b\5\2\65\66\7\7\2\2\66\67\5\16\b\2\67\21")
        buf.write("\3\2\2\2\6\24\33\'\61")
        return buf.getvalue()


class CFGParser ( Parser ):

    grammarFileName = "CFG.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "';'", "'1'", "'.'", "'|'", "'->'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "IDENTIFIER", "WS" ]

    RULE_context_free_grammar = 0
    RULE_rule_list = 1
    RULE_symbol = 2
    RULE_variable = 3
    RULE_symbol_list = 4
    RULE_alternative = 5
    RULE_alternative_list = 6
    RULE_rule_ = 7

    ruleNames =  [ "context_free_grammar", "rule_list", "symbol", "variable", 
                   "symbol_list", "alternative", "alternative_list", "rule_" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    IDENTIFIER=6
    WS=7

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.8")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class Context_free_grammarContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def rule_list(self):
            return self.getTypedRuleContext(CFGParser.Rule_listContext,0)


        def getRuleIndex(self):
            return CFGParser.RULE_context_free_grammar

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitContext_free_grammar" ):
                return visitor.visitContext_free_grammar(self)
            else:
                return visitor.visitChildren(self)




    def context_free_grammar(self):

        localctx = CFGParser.Context_free_grammarContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_context_free_grammar)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 16
            self.rule_list()
            self.state = 18
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==CFGParser.T__0:
                self.state = 17
                self.match(CFGParser.T__0)


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Rule_listContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def rule_(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(CFGParser.Rule_Context)
            else:
                return self.getTypedRuleContext(CFGParser.Rule_Context,i)


        def getRuleIndex(self):
            return CFGParser.RULE_rule_list

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitRule_list" ):
                return visitor.visitRule_list(self)
            else:
                return visitor.visitChildren(self)




    def rule_list(self):

        localctx = CFGParser.Rule_listContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_rule_list)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 20
            self.rule_()
            self.state = 25
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,1,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    self.state = 21
                    self.match(CFGParser.T__0)
                    self.state = 22
                    self.rule_() 
                self.state = 27
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,1,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class SymbolContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self):
            return self.getToken(CFGParser.IDENTIFIER, 0)

        def getRuleIndex(self):
            return CFGParser.RULE_symbol

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitSymbol" ):
                return visitor.visitSymbol(self)
            else:
                return visitor.visitChildren(self)




    def symbol(self):

        localctx = CFGParser.SymbolContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_symbol)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 28
            _la = self._input.LA(1)
            if not(_la==CFGParser.T__1 or _la==CFGParser.IDENTIFIER):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class VariableContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self):
            return self.getToken(CFGParser.IDENTIFIER, 0)

        def getRuleIndex(self):
            return CFGParser.RULE_variable

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitVariable" ):
                return visitor.visitVariable(self)
            else:
                return visitor.visitChildren(self)




    def variable(self):

        localctx = CFGParser.VariableContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_variable)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 30
            self.match(CFGParser.IDENTIFIER)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Symbol_listContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def symbol(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(CFGParser.SymbolContext)
            else:
                return self.getTypedRuleContext(CFGParser.SymbolContext,i)


        def getRuleIndex(self):
            return CFGParser.RULE_symbol_list

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitSymbol_list" ):
                return visitor.visitSymbol_list(self)
            else:
                return visitor.visitChildren(self)




    def symbol_list(self):

        localctx = CFGParser.Symbol_listContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_symbol_list)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 32
            self.symbol()
            self.state = 37
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==CFGParser.T__2:
                self.state = 33
                self.match(CFGParser.T__2)
                self.state = 34
                self.symbol()
                self.state = 39
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class AlternativeContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def symbol_list(self):
            return self.getTypedRuleContext(CFGParser.Symbol_listContext,0)


        def getRuleIndex(self):
            return CFGParser.RULE_alternative

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitAlternative" ):
                return visitor.visitAlternative(self)
            else:
                return visitor.visitChildren(self)




    def alternative(self):

        localctx = CFGParser.AlternativeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_alternative)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 40
            self.symbol_list()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Alternative_listContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def alternative(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(CFGParser.AlternativeContext)
            else:
                return self.getTypedRuleContext(CFGParser.AlternativeContext,i)


        def getRuleIndex(self):
            return CFGParser.RULE_alternative_list

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitAlternative_list" ):
                return visitor.visitAlternative_list(self)
            else:
                return visitor.visitChildren(self)




    def alternative_list(self):

        localctx = CFGParser.Alternative_listContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_alternative_list)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 42
            self.alternative()
            self.state = 47
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==CFGParser.T__3:
                self.state = 43
                self.match(CFGParser.T__3)
                self.state = 44
                self.alternative()
                self.state = 49
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Rule_Context(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def variable(self):
            return self.getTypedRuleContext(CFGParser.VariableContext,0)


        def alternative_list(self):
            return self.getTypedRuleContext(CFGParser.Alternative_listContext,0)


        def getRuleIndex(self):
            return CFGParser.RULE_rule_

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitRule_" ):
                return visitor.visitRule_(self)
            else:
                return visitor.visitChildren(self)




    def rule_(self):

        localctx = CFGParser.Rule_Context(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_rule_)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 50
            self.variable()
            self.state = 51
            self.match(CFGParser.T__4)
            self.state = 52
            self.alternative_list()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import Iterable


def header(msg):
    print('-------------------')
    print('--- {}'.format(msg))
    print('-------------------')


def print_words(words: Iterable[str], epsilon='ε', separator=', ', leftsep='{', rightsep='}') -> str:
    words = sorted([w if w not in ['', epsilon] else 'ε' for w in words])
    return '{}{}{}'.format(leftsep, separator.join(sorted(words, key=lambda x: (len(x), x if x != 'ε' else ''))), rightsep)



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

import codecs


def read_text(filename):
    with open(filename, 'r') as f:
        return f.read()


def write_text(filename, text):
    with open(filename, 'w') as f:
        f.write(text)


def read_utf8_text(filename):
    with codecs.open(filename,'r',encoding='utf8') as f:
        return f.read()


def write_utf8_text(filename, text):
    with codecs.open(filename,'w',encoding='utf8') as f:
        f.write(text)


def remove_comments(text: str) -> str:
    lines = []
    for line in text.split('\n'):
        line = line.strip()
        if line and not line.startswith('%'):
            lines.append(line)
    return '\n'.join(lines)



# Generated from CFG.g4 by ANTLR 4.8
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .CFGParser import CFGParser
else:
    from CFGParser import CFGParser

# This class defines a complete generic visitor for a parse tree produced by CFGParser.

class CFGVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by CFGParser#context_free_grammar.
    def visitContext_free_grammar(self, ctx:CFGParser.Context_free_grammarContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CFGParser#rule_list.
    def visitRule_list(self, ctx:CFGParser.Rule_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CFGParser#symbol.
    def visitSymbol(self, ctx:CFGParser.SymbolContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CFGParser#variable.
    def visitVariable(self, ctx:CFGParser.VariableContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CFGParser#symbol_list.
    def visitSymbol_list(self, ctx:CFGParser.Symbol_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CFGParser#alternative.
    def visitAlternative(self, ctx:CFGParser.AlternativeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CFGParser#alternative_list.
    def visitAlternative_list(self, ctx:CFGParser.Alternative_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by CFGParser#rule_.
    def visitRule_(self, ctx:CFGParser.Rule_Context):
        return self.visitChildren(ctx)



del CFGParser


#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

import copy
from typing import List

from gambatools.cfg import CFG, Variable
from gambatools.cfg_algorithms import parse_simple_cfg, cfg_add_new_start_variable_in_place, \
    cfg_remove_epsilon_rules_in_place, cfg_eliminate_unit_rules_in_place, cfg_make_rules_of_length_two_in_place, \
    cfg_eliminate_terminals_in_place, cfg_to_chomsky, cfg_words_up_to_n
from gambatools.language_generator import compare_languages
from gambatools.notebook import print_feedback


def check_cfg_has_start_variable(G: CFG, S: Variable) -> List[str]:
    if G.S != S:
        return ['Error: the start variable should be {}'.format(S)]
    return []


def check_cfg_has_no_epsilon_rules(G: CFG) -> List[str]:
    for r in G.R:
        if r.is_epsilon() and r.variable != G.S:
            return ['Error: the CFG has an epsilon rule: {}'.format(r)]
    return []


def check_cfg_has_no_unit_productions(G: CFG) -> List[str]:
    for r in G.R:
        if r.is_unit_rule():
            return ['Error: the CFG has a unit rule: {}'.format(r)]
    return []


def check_cfg_has_right_hand_sides_of_length_at_most_two(G: CFG) -> List[str]:
    for r in G.R:
        if len(r.alternative.symbols) > 2:
            return ['Error: the CFG has a rule with more than two symbols: {}'.format(r)]
    return []


def check_cfg_is_chomsky(G: CFG) -> List[str]:
    for r in G.R:
        if not r.is_chomsky():
            return ['Error: the CFG has a rule that is not in Chomsky format: {}'.format(r)]
    return []


def cfg_apply_chomsky(G: CFG, phase: int, start_variable: str) -> CFG:
    G1 = copy.deepcopy(G)
    if phase >= 1:
        cfg_add_new_start_variable_in_place(G1, start_variable)
    if phase >= 2:
        cfg_remove_epsilon_rules_in_place(G1)
    if phase >= 3:
        cfg_eliminate_unit_rules_in_place(G1)
    if phase >= 4:
        cfg_make_rules_of_length_two_in_place(G1)
    if phase >= 5:
        cfg_eliminate_terminals_in_place(G1)
    return G1


def cfg_check_chomsky(cfg: str, cfg1: str, phase: int, start_variable: str, length: int) -> None:
    try:
        feedback = []
        G = parse_simple_cfg(cfg)
        G1 = parse_simple_cfg(cfg1)
        S = Variable(start_variable)

        if phase >= 0:
            g = cfg_to_chomsky(G) if not G.is_chomsky() else G
            g1 = cfg_to_chomsky(G1) if not G1.is_chomsky() else G1
            A1 = cfg_words_up_to_n(g1, length)
            A2 = cfg_words_up_to_n(g, length)
            feedback = feedback + compare_languages(A1, A2)
        if phase >= 1:
            feedback = feedback + check_cfg_has_start_variable(G1, S)
        if phase >= 2:
            feedback = feedback + check_cfg_has_no_epsilon_rules(G1)
        if phase >= 3:
            feedback = feedback + check_cfg_has_no_unit_productions(G1)
        if phase >= 4:
            feedback = feedback + check_cfg_has_right_hand_sides_of_length_at_most_two(G1)
        if phase >= 5:
            feedback = feedback + check_cfg_is_chomsky(G1)
        print_feedback(feedback)
    except Exception as e:
        print('Error: {}'.format(e))



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

#!/usr/bin/env python3

from gambatools.regexp import *
from gambatools.regexpParser import *
from gambatools.regexpVisitor import *
from gambatools.regexpLexer import *


def concatenation(expressions):
    if len(expressions) == 1:
        return expressions[0]
    return Concat(expressions[0], concatenation(expressions[1:]))


class regexpVisitor(regexpVisitor):

    def __init__(self):
        super(regexpVisitor, self).__init__()

    def make_unary_expression(self, unaryop, ctx):
        operand = self.visit(ctx.expression())
        return unaryop(operand)

    def make_binary_expression(self, binaryop, ctx):
        left = self.visit(ctx.expression(0))
        right = self.visit(ctx.expression(1))
        return binaryop(left, right)

    def visitOneExpression(self, ctx: regexpParser.OneExpressionContext):
        return One()

    def visitSymbolExpression(self, ctx: regexpParser.SymbolExpressionContext):
        return Symbol(ctx.getText())

    def visitIterationExpression(self, ctx: regexpParser.IterationExpressionContext):
        return self.make_unary_expression(Iteration, ctx)

    def visitConcatExpression(self, ctx: regexpParser.ConcatExpressionContext):
        return self.make_binary_expression(Concat, ctx)

    def visitSumExpression(self, ctx: regexpParser.SumExpressionContext):
        return self.make_binary_expression(Sum, ctx)

    def visitZeroExpression(self, ctx: regexpParser.ZeroExpressionContext):
        return Zero()

    def visitParensExpression(self, ctx: regexpParser.ParensExpressionContext):
        return self.visit(ctx.expression())


def parse_regexp(text):
    lexer = regexpLexer(InputStream(text))
    stream = CommonTokenStream(lexer)
    parser = regexpParser(stream)
    tree = parser.expression()
    visitor = regexpVisitor()
    return visitor.visit(tree)



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from collections import defaultdict
from typing import Set, Mapping, MutableMapping, Tuple, Union, List, Optional
import io

from gambatools.automaton import Automaton
from gambatools.automaton_algorithms import default_transition_label_regex, default_state_label_regex, AutomatonParser, AutomatonBuilder, \
    nfa_keywords
from gambatools.dfa import State, Symbol, print_state_set, DFA
from gambatools.nfa import NFA
from gambatools.identifier_generator import IdentifierGenerator


def _nfa_cache(N: NFA) -> Tuple[Mapping[State, Set[State]], Mapping[Tuple[State, Symbol], Set[State]]]:
    # make a cache of epsilon_closure(q)
    Eq: MutableMapping[State, Set[State]] = {}
    for q in N.Q:
        Eq[q] = epsilon_closure(N, q)

    # make a cache of epsilon_closure(delta(q,a))
    Eqa = defaultdict(set)  # Eqa: TypedDict[State, Set[str]]
    for (q, a), Q in N.delta.items():
        Eqa[(q, a)] = set().union(*[Eq[q] for q in Q]) if Q else set([])

    return Eq, Eqa


def epsilon_closure(N: NFA, q: Union[State, Set[State]]) -> Set[State]:
    result: Set[State] = q.copy() if isinstance(q, set) else {q}
    todo: Set[State] = result.copy()
    while todo:
        q = todo.pop()
        Q1: Set[State] = N.delta[q, N.epsilon] - result
        result = result | Q1
        todo = todo | Q1
    return result


def nfa_accepts_word(N: NFA, word: str) -> bool:
    q0 = N.q0
    F = N.F
    Eq, Eqa = _nfa_cache(N)
    q: Set[State] = Eq[q0]
    for a in word:
        q = set().union(*[Eqa[q_i, Symbol(a)] for q_i in q])
    return not q.isdisjoint(F)


def nfa_words_up_to_n(N: NFA, n: int) -> Set[str]:
    Eq, Eqa = _nfa_cache(N)

    # F contains all states that can terminate
    F1 = [q for q in N.Q if not Eq[q].isdisjoint(N.F)]

    result = set([])
    if N.q0 in F1:
        result.add('')

    W = defaultdict(lambda: set([]))  # W: TypedDict[State, Set[str]]
    for q in Eq[N.q0]:
        W[q] = {''}

    for i in range(n):
        W1 = defaultdict(lambda: set([]))  # W1: TypedDict[State, Set[str]]
        for (q, words) in W.items():
            for a in N.Sigma:
                if (q, a) in Eqa:
                    for q1 in Eqa[(q, a)]:
                        words_q1 = set([word + a for word in words])
                        W1[q1] |= words_q1
                        if q1 in F1:
                            result |= words_q1
        W = W1
    return result


def nfa_do_transition(N: NFA, a: Symbol, R: Set[State]) -> Set[State]:
    """Returns all NFA states reachable from an element in R via an a-transition."""
    delta = N.delta
    result: Set[State] = set([])
    for r in R:
        if (r, a) in delta:
            result |= delta[r, a]
    return result

def nfa_find_epsilon_path(N: NFA, R: Set[State], f: State) -> Optional[List[State]]:
    """Returns a path from an element r in R to f"""
    delta = N.delta
    epsilon = N.epsilon

    if f in R:
        return [f]

    backpointers = {}

    def make_path(q: State) -> List[State]:
        path = [q]
        while q not in R:
            q = backpointers[q]
            path.insert(0, q)
        return path

    visited: Set[State] = set([r for r in R])
    todo: Set[State] = set([r for r in R])
    while len(todo) > 0:
        src = todo.pop()
        for (p, a), Q1 in delta.items():
            if p != src or a != epsilon:
                continue
            for q in Q1:
                target = q
                backpointers[target] = src
                if target == f:
                    return make_path(target)
                if target not in visited:
                    todo.add(target)
                    visited.add(target)
    return None


def nfa_find_transition(N: NFA, R: Set[State], a: Symbol, target: State) -> Optional[State]:
    """Returns src such that src --a--> target"""
    delta = N.delta
    for src in R:
        for (p, a1), Q1 in delta.items():
            if src != p or a != a1:
                continue
            for q in Q1:
                if q == target:
                    return src
    return None


def nfa_simulate_word(N: NFA, w: str) -> Optional[Tuple[State, str]]:
    F = N.F
    H = []

    R = {N.q0}
    H.append(R)
    R = epsilon_closure(N, R)
    H.append(R)

    for a in w:
        R = nfa_do_transition(N, Symbol(a), R)
        H.append(R)
        R = epsilon_closure(N, R)
        H.append(R)

    if any(r in F for r in R):
        S = H.pop()
        front = next(r for r in S if r in F)
        result = [(front, '')]
        word = ''
        for a in reversed(w):
            S = H.pop()
            path = nfa_find_epsilon_path(N, S, front)
            result = [(r, word) for r in path[:-1]] + result
            front = path[0]
            S = H.pop()
            front = nfa_find_transition(N, S, a, front)
            word = a + word
            result = [(front, word)] + result
        S = H.pop()
        path = nfa_find_epsilon_path(N, S, front)
        result = [(r, word) for r in path[:-1]] + result
        return result
    else:
        return None


def nfa_to_dfa(N: NFA) -> DFA:
    def state(q: Set[State]) -> State:
        return State(print_state_set(q))

    Sigma: Set[Symbol] = N.Sigma.copy()
    F: Set[State] = set([])
    Q0: Set[State] = epsilon_closure(N, {N.q0})
    stateQ0 = state(Q0)
    Q: Set[State] = {stateQ0}
    delta: MutableMapping[Tuple[State, Symbol], State] = {}
    if not Q0.isdisjoint(N.F):
        F.add(stateQ0)
    todo = [Q0]

    while todo:
        Q1 = todo.pop()
        stateQ1 = state(Q1)

        for a in Sigma:
            Q2 = set([])
            for q1 in Q1:
                Q2 |= N.delta[q1, a]
            Q2 = epsilon_closure(N, Q2)
            stateQ2 = state(Q2)
            delta[stateQ1, a] = stateQ2
            if not Q2.isdisjoint(N.F):
                F.add(stateQ2)
            if stateQ2 not in Q:
                Q.add(stateQ2)
                todo.append(Q2)

    return DFA(Q, Sigma, delta, stateQ0, F)


def random_nfa(Sigma: Set[Symbol], n: int) -> NFA:
    import random

    Q: Set[State] = set([State('q{}'.format(i)) for i in range(n)])
    delta: MutableMapping[Tuple[State, Symbol], Set[State]] = {}
    q0: State = next(q for q in Q if q == State('q0'))
    F: Set[State] = set([])
    epsilon = Symbol('')

    # generate delta
    for q in Q:
        for a in Sigma | {epsilon}:
            k = random.choice(range(min(n + 1, 3)))
            Q1 = set(random.sample(list(Q), k))
            delta[q, a] = Q1

    # generate F
    for q in Q:
        if random.randint(0, 5) <= 2:
            F.add(q)
    return NFA(Q, Sigma, delta, q0, F, epsilon)


def nfa_repetition(N: NFA, id_generator: IdentifierGenerator = IdentifierGenerator()) -> NFA:
    Sigma = N.Sigma
    q0 = State(id_generator.generate('q'))
    Q = N.Q | {q0}
    F = N.F | {q0}
    delta = defaultdict(lambda: set([]))
    delta.update(N.delta)
    for q in F:
        delta[q, N.epsilon] |= {N.q0}
    delta[q0, N.epsilon] = {N.q0}
    return NFA(Q, Sigma, delta, q0, F)


def nfa_union(N1: NFA, N2: NFA, id_generator: IdentifierGenerator = IdentifierGenerator()) -> NFA:
    assert N1.Q.isdisjoint(N2.Q)
    Sigma = N1.Sigma | N2.Sigma
    q0 = State(id_generator.generate('q'))
    Q = N1.Q | N2.Q | {q0}
    F = N1.F | N2.F
    delta = defaultdict(lambda: set([]))
    delta.update(N1.delta)
    delta.update(N2.delta)
    delta[q0, N1.epsilon] = {N1.q0, N2.q0}
    return NFA(Q, Sigma, delta, q0, F)


def nfa_concatenation(N1: NFA, N2: NFA) -> NFA:
    assert N1.Q.isdisjoint(N2.Q)
    Sigma = N1.Sigma | N2.Sigma
    q0 = N1.q0
    Q = N1.Q | N2.Q | {q0}
    F = N2.F
    delta = defaultdict(lambda: set([]))
    delta.update(N1.delta)
    delta.update(N2.delta)
    for q in N1.F:
        delta[q, N1.epsilon] |= {N2.q0}
    return NFA(Q, Sigma, delta, q0, F)


def print_nfa(N: NFA) -> str:
    Q = N.Q
    Sigma = N.Sigma
    delta = N.delta
    q0 = N.q0
    F = N.F
    epsilon = N.epsilon

    out = io.StringIO()
    out.write('states {}\n'.format(' '.join(sorted(Q))))
    out.write('final {}\n'.format(' '.join(sorted(F))))
    out.write('initial {}\n'.format(q0))
    out.write('input_symbols {}\n'.format(' '.join(sorted(Sigma))))
    out.write('epsilon {}\n'.format(epsilon))
    transitions = defaultdict(lambda: [])
    for (p, a), Q1 in delta.items():
        for q in Q1:
            transitions['{} {}'.format(p, q)].append(a)
    for pq in sorted(transitions.keys()):
        out.write('{} {}\n'.format(pq, ' '.join(transitions[pq])))
    result = out.getvalue()
    out.close()
    return result


def automaton_to_nfa(A: Automaton, transition_regex = default_transition_label_regex(), state_regex = default_state_label_regex()) -> NFA:
    return NFABuilder(A, state_regex=state_regex, transition_regex=transition_regex).build()


def parse_nfa(text: str, transition_regex = default_transition_label_regex(), state_regex = default_state_label_regex()) -> NFA:
    A = AutomatonParser(state_regex=state_regex, transition_regex=transition_regex, keywords = nfa_keywords()).parse(text)
    return automaton_to_nfa(A, state_regex=state_regex, transition_regex=transition_regex)


class NFABuilder(AutomatonBuilder):
    def __init__(self, A: Automaton, state_regex=r'\w+', transition_regex=r'\w+'):
        super().__init__(A, state_regex, transition_regex)

    def used_input_symbols(self, epsilon: str) -> Set[str]:
        A = self.A
        return set(a for (_, a, _) in A.transitions if a != epsilon)

    def build(self) -> NFA:
        A = self.A
        A.states = A.states if len(A.states) > 0 else A.used_states()

        self._check_states_are_declared()
        self._check_state_labels()
        self._check_one_initial_state()
        epsilon = self.parse_symbol()
        input_symbols = self.get_symbol_set('input_symbols', self.used_input_symbols(epsilon))
        self._check_symbols(input_symbols)

        Q = set(State(s) for s in A.states)
        Sigma = set(Symbol(s) for s in input_symbols)
        delta = defaultdict(set)
        q0 = State(next(iter(A.initial_states)))
        F = set(State(s) for s in A.final_states)
        epsilon = Symbol(epsilon)
        for (p, a, q) in A.transitions:
            p = State(p)
            q = State(q)
            a = Symbol(a)
            delta[p, a].add(q)
        return NFA(Q, Sigma, delta, q0, F, epsilon)



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from typing import List


def remove_none(seq: List):
    return [x for x in seq if x]


def remove_duplicates(seq: List):
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]


def remove_if(seq: List, predicate):
    index = len(seq) - 1
    while index >= 0:
        if predicate(seq[index]):
            del seq[index]
        index = index - 1
    return seq



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

import copy
import itertools
import re
import string
from collections import defaultdict
from typing import List, Set, MutableMapping, Tuple, Union, Optional, DefaultDict

from gambatools.algorithms import last_index, first_index
from gambatools.dfa import State, Symbol, DFA
from gambatools.nfa import NFA
from gambatools.cfg import CFG, Terminal, Variable, Alternative, Rule, DerivationTerm
from gambatools.list_utility import remove_none, remove_duplicates, remove_if


class BaetenCFGParser(object):

    @staticmethod
    def is_comment(line: str):
        return line.strip().startswith('#')

    @staticmethod
    def parse_variable(ch: str) -> Union[Terminal, Variable]:
        return Terminal(ch) if ch.islower() else Variable(ch)

    def parse_alternative(self, text: str) -> Optional[Alternative]:
        if text == '0':
            return None
        elif text == '1':
            return Alternative([])
        return Alternative([self.parse_variable(ch) for ch in text])

    def parse_rule(self, text: str) -> List[Rule]:
        if self.is_comment(text):
            return []
        words = list(filter(None, re.split(r'\s+', text)))
        if len(words) < 2:
            return []
        lhs = Variable(words[0])
        alternatives = remove_none([self.parse_alternative(alternative) for alternative in words[1:]])
        return [Rule(lhs, alternative) for alternative in alternatives]

    def parse_grammar(self, text: str, check_validity: bool = True):
        text = re.sub('[=+]', '', text)
        R = list(itertools.chain.from_iterable(self.parse_rule(line) for line in text.split('\n')))
        V = set([rule.variable for rule in R])
        Sigma = set().union(*[rule.terminals() for rule in R])
        S = R[0].variable
        return CFG(V, Sigma, R, S, check_validity=check_validity)


class SimpleCFGParser(object):
    def __init__(self):
        def repeat(x):
            return '(' + x + r')*'
        self.epsilon = None
        self.zero = '@'
        self.alternative_regex = r'[\.\w]+'
        self.variable_regex = r'\w+'
        S = r'\s*'
        A = self.alternative_regex
        V = self.variable_regex
        bar = r'\|'
        self.rule_regex = S + V + S + '->' + S + A + repeat(S + bar + S + A + S)

    @staticmethod
    def is_comment(line: str):
        return line.strip().startswith('%')

    @staticmethod
    def parse_variable(ch: str) -> Union[Terminal, Variable]:
        return Terminal(ch) if ch.islower() else Variable(ch)

    def parse_alternative(self, text: str) -> Optional[Alternative]:
        if text == self.zero:
            return None
        elif text == self.epsilon:
            return Alternative([])
        return Alternative([self.parse_variable(ch) for ch in text])

    def parse_rule(self, line: str) -> List[Rule]:
        words = re.split(r'->', line)
        if len(words) != 2:
            raise RuntimeError('invalid production {}'.format(line))
        lhs = Variable(words[0].strip())
        alternatives = words[1].strip().split('|')
        alternatives = remove_none([self.parse_alternative(alternative.strip()) for alternative in alternatives])
        return [Rule(lhs, alternative) for alternative in alternatives]

    def parse_epsilon(self, lines):
        result = []
        for line in lines:
            line = line.strip()
            if not line or line.startswith('%'):
                pass
            elif line.startswith('epsilon'):
                m = re.fullmatch(r'epsilon\s*=\s*(\w)', line)
                if not m:
                    raise RuntimeError('invalid epsilon declaration {}'.format(line))
                self.epsilon = m.group(1)
            else:
                if re.fullmatch(self.rule_regex, line):
                    result.append(line)
                else:
                    raise RuntimeError('invalid production {}'.format(line))
        if not self.epsilon:
            self.epsilon = 'ε' if any(['ε' in line for line in result]) else '_'
        return result

    def parse_grammar(self, text: str, check_validity: bool = True):
        lines = text.split('\n')
        lines = self.parse_epsilon(lines)
        R = list(itertools.chain.from_iterable(self.parse_rule(line) for line in lines))
        if not R:
            raise RuntimeError('the grammar has no rules')
        V = set([rule.variable for rule in R])
        Sigma = set().union(*[rule.terminals() for rule in R])
        S = R[0].variable
        return CFG(V, Sigma, R, S, epsilon=Terminal(self.epsilon), check_validity=check_validity)


def parse_cfg_baeten(text: str, check_validity: bool = True) -> CFG:
    """
    Parse a CFG in the format used by Baeten:

    S = aA + bB
    A = a + B
    B = b + 1

    Variables are upper case, 1 is the empty string.
    :param text: a string containing a grammar
    :param check_validity:
    :return: the grammar
    """

    parser = BaetenCFGParser()
    return parser.parse_grammar(text, check_validity)


def parse_simple_cfg(text: str, check_validity: bool = True) -> CFG:
    """
    Parse a CFG in a format used by Sipser:

    S -> aA | bB
    A -> a | B
    B -> b | _

    Variables are upper case, _ is the empty string.
    :param text: a string containing a grammar
    :param check_validity:
    :return: the grammar
    """

    parser = SimpleCFGParser()
    return parser.parse_grammar(text, check_validity)


def cfg_to_dfa(G: CFG, check_validity: bool = True) -> DFA:
    def is_epsilon(alternative: Alternative):
        return len(alternative.symbols) == 0 or (len(alternative.symbols) == 1 and alternative.symbols[0] == G.epsilon)

    def is_transition(alternative: Alternative) -> bool:
        return len(alternative.symbols) == 2 and isinstance(alternative.symbols[0], Terminal) and isinstance(alternative.symbols[1], Variable)

    Sigma: Set[Symbol] = set([Symbol(t) for t in G.Sigma])
    Q: Set[State] = set([State(t) for t in G.V])
    delta: MutableMapping[Tuple[State, Symbol], State] = {}
    F: Set[State] = set([])
    q0: State = next(q for q in Q if q == State(G.S))

    for rule in G.R:
        q = State(rule.variable)
        if is_epsilon(rule.alternative):
            F.add(q)
        elif is_transition(rule.alternative):
            a = Symbol(rule.alternative.symbols[0])
            q1 = State(rule.alternative.symbols[1])
            delta[q, a] = q1
        else:
            raise RuntimeError('cga2nfa: alternative {} has the wrong format'.format(rule.alternative))
    return DFA(Q, Sigma, delta, q0, F, check_validity=check_validity)


def cfg_to_nfa(grammar: CFG) -> NFA:
    """Transforms a CFG to an NFA. The CFG must already be in the right format."""
    def is_epsilon(alternative: Alternative):
        return len(alternative.symbols) == 0 or (len(alternative.symbols) == 1 and alternative.symbols[0] == grammar.epsilon)

    def is_epsilon_transition(alternative: Alternative) -> bool:
        return len(alternative.symbols) == 1 and isinstance(alternative.symbols[0], Variable)

    def is_transition(alternative: Alternative) -> bool:
        return len(alternative.symbols) == 2 and isinstance(alternative.symbols[0], Terminal) and isinstance(alternative.symbols[1], Variable)

    Sigma: Set[Symbol] = set([Symbol(t) for t in grammar.Sigma])
    Q: Set[State] = set([State(t) for t in grammar.V])
    delta = defaultdict(lambda: set([]))  # delta: TypedDict[Tuple[State, Symbol], Set[State]]
    F: Set[State] = set([])
    q0: State = next(q for q in Q if q == State(grammar.S))
    epsilon = Symbol('')

    for rule in grammar.R:
        q = State(rule.variable)
        if is_epsilon(rule.alternative):
            F.add(q)
        elif is_epsilon_transition(rule.alternative):
            q1 = State(rule.alternative.symbols[0])
            delta[q, epsilon].add(q1)
        elif is_transition(rule.alternative):
            a = Symbol(rule.alternative.symbols[0])
            q1 = State(rule.alternative.symbols[1])
            delta[q, a].add(q1)
        else:
            raise RuntimeError('cga2nfa: alternative {} has the wrong format'.format(rule.alternative))
    return NFA(Q, Sigma, delta, q0, F, epsilon)


def expand_nullable_variables(x: DerivationTerm, W: Set[Variable]) -> List[DerivationTerm]:
    """Expands nullable variables (elements in W) in x."""
    if len(x) == 0:
        return [x]
    y = expand_nullable_variables(x[1:], W)
    result = [[x[0]] + y_i for y_i in y]
    if isinstance(x[0], Variable) and x[0] in W:
        result = result + y
    return result


def cfg_nullable_variables(G: CFG) -> Set[Variable]:
    R = G.R
    nullable = set([])

    while True:
        changed = False
        for r in R:
            if r.variable not in nullable and all(x in nullable for x in r.alternative.symbols):
                nullable.add(r.variable)
                changed = True
        if not changed:
            break

    return nullable


def cfg_remove_epsilon_rules_in_place(G: CFG) -> None:
    R = G.R
    S = G.S

    W = cfg_nullable_variables(G)

    # R1 contains the new list of rules
    R1 = []

    for rule in R:
        for symbols in expand_nullable_variables(rule.alternative.symbols, W):
            if not symbols and rule.variable in W - {S}:
                continue
            R1.append(Rule(rule.variable, Alternative(symbols)))

    R1 = remove_duplicates(R1)
    R.clear()
    R.extend(R1)


def cfg_remove_epsilon_rules(G: CFG) -> CFG:
    G = copy.deepcopy(G)
    cfg_remove_epsilon_rules_in_place(G)
    return G


def cfg_derivable_variables(G: CFG, A: Variable) -> Set[Variable]:
    R = G.R
    V = G.V

    W: Set[Variable] = set([])
    W1: Set[Variable] = set([])

    for r in R:
        if r.variable == A and len(r.alternative.symbols) == 1:
            B = r.alternative.symbols[0]
            if B in V:
                W.add(B)

    while W1 != W:
        W1 = W.copy()
        for r in R:
            if len(r.alternative.symbols) == 1:
                C = r.variable
                B = r.alternative.symbols[0]
                if C in W1 and B in V:
                    W.add(B)
    return W - {A}


def cfg_put_start_variable_in_front(G: CFG) -> None:
    R = G.R
    S = G.S
    for i in range(len(R)):
        if R[i].variable == S:
            R[0], R[i] = R[i], R[0]
            break


def cfg_eliminate_unit_rules_in_place(G: CFG) -> None:
    R = G.R
    V = G.V
    R1 = R.copy()

    for A in V:
        W = cfg_derivable_variables(G, A)
        for r in R:
            B = r.variable
            if B in W and not r.is_unit_rule():
                alpha = r.alternative
                r1 = Rule(A, alpha)
                if not r1 in R1:
                    R1.append(r1)
    R1 = remove_if(R1, lambda x: x.is_unit_rule())
    G.R = R1
    cfg_put_start_variable_in_front(G)


def cfg_eliminate_unit_rules(G: CFG) -> CFG:
    G = copy.deepcopy(G)
    cfg_eliminate_unit_rules_in_place(G)
    return G


def cfg_fresh_variable(G: CFG, hint: str) -> Variable:
    # assert len(hint) == 1
    # assert hint.isupper()

    V = G.V
    if len(V) >= 26:
        index = 0
        A = Variable(hint)
        while A in V:
            A = Variable('{}{}'.format(hint, index))
            index = index + 1
        return A
    else:
        if Variable(hint) not in V:
            A = Variable(hint)
            return A
        for A in string.ascii_uppercase:
            if Variable(A) not in V:
                A = Variable(A)
                return A


def cfg_add_new_start_variable_in_place(G: CFG, hint: str = 'S') -> None:
    S0 = cfg_fresh_variable(G, hint)
    G.R.insert(0, Rule(S0, Alternative([G.S])))
    G.V.add(S0)
    G.S = S0


def cfg_add_new_start_variable(G: CFG, hint: str = 'S') -> CFG:
    G = copy.deepcopy(G)
    cfg_add_new_start_variable_in_place(G, hint)
    return G


def cfg_make_rules_of_length_two_in_place(G: CFG) -> None:
    def fresh_variables(hint: str, n: int) -> List[Variable]:
        """Returns n fresh variables and adds them to G.V"""
        result = []
        index = 0
        for i in range(n):
            P = cfg_fresh_variable(G, hint)
            result.append(P)
            G.V.add(P)
        return result

    for rule in G.R.copy():
        A = rule.variable
        a = rule.alternative
        u = a.symbols
        n = len(u)
        if n <= 2:
            continue
        A = fresh_variables(A, n - 2)
        for k in range(n - 3):
            G.R.append(Rule(A[k], Alternative([u[k + 1], A[k + 1]])))
        G.R.append(Rule(A[-1], Alternative(u[-2:])))
        a.symbols = [u[0], A[0]]


def cfg_make_rules_of_length_two(G: CFG) -> CFG:
    G = copy.deepcopy(G)
    cfg_make_rules_of_length_two_in_place(G)
    return G


def cfg_eliminate_terminals_in_place(G: CFG) -> None:
    replacements: MutableMapping[Terminal, Variable] = {}

    def replace_symbol(symbol: Union[Variable, Terminal]) -> Variable:
        if isinstance(symbol, Terminal):
            if symbol not in replacements:
                A = cfg_fresh_variable(G, symbol.upper())
                replacements[symbol] = A
                G.V.add(A)
            return replacements[symbol]
        else:
            return symbol

    for rule in G.R:
        a = rule.alternative
        if len(a.symbols) >= 2:
            a.symbols = [replace_symbol(symbol) for symbol in a.symbols]

    for terminal, variable in replacements.items():
        A = variable
        a = Alternative([terminal])
        G.R.append(Rule(A, a))


def cfg_eliminate_terminals(G: CFG) -> CFG:
    G = copy.deepcopy(G)
    cfg_eliminate_terminals_in_place(G)
    return G


def cfg_to_chomsky_in_place(G: CFG, verbose: bool = False) -> None:
    if verbose: print('--- start ---\n', G, ' V =', *G.V)
    cfg_add_new_start_variable_in_place(G)
    if verbose: print('--- cfg_add_new_start_variable_in_place ---\n', G, '\n variables =', *G.V)
    cfg_remove_epsilon_rules_in_place(G)
    if verbose: print('--- cfg_remove_epsilon_rules_in_place ---\n', G, '\n variables =', *G.V)
    cfg_eliminate_unit_rules_in_place(G)
    if verbose: print('--- cfg_eliminate_unit_rules_in_place ---\n', G, '\n variables =', *G.V)
    cfg_make_rules_of_length_two_in_place(G)
    if verbose: print('--- cfg_make_rules_of_length_two_in_place ---\n', G, '\n variables =', *G.V)
    cfg_eliminate_terminals_in_place(G)
    if verbose: print('--- cfg_eliminate_terminals_in_place ---\n', G, '\n variables =', *G.V)


def cfg_to_chomsky(G: CFG, verbose: bool = False) -> CFG:
    G = copy.deepcopy(G)
    cfg_to_chomsky_in_place(G, verbose)
    return G


def cfg_cyk_matrix(G: CFG, w: str, verbose: bool = False) -> DefaultDict[Tuple[int, int], Set[Variable]]:
    assert G.is_chomsky()

    V = G.V
    R = G.R
    n = len(w)

    # Build a mapping P of all productions
    P = defaultdict(lambda: [])
    for rule in R:
        P[rule.variable].append(rule.alternative.symbols)

    # X[i, j] will contain { A | A -> w[i].w[i+1]...w[j] }
    X = defaultdict(lambda: set([]))
    for i in range(n):
        w_i = [Terminal(w[i])]
        X[i, i] = set(A for A in V if w_i in P[A])

    if verbose:
        print('--- start ---')
        print('V = ', *V)
        for i, j in X:
            print('X[{}, {}] = {}'.format(i, j, X[i, j]))

    for m in range(1, n):
        for i in range(n - m):
            j = i + m
            if verbose: print('processing X[{}, {}]'.format(i, j))
            for k in range(i, j):
                if verbose: print('X[{}, {}] depends on X[{}, {}] and X[{}, {}]'.format(i, j, i, k, k + 1, j))
                for (B, C) in itertools.product(X[i, k], X[k + 1, j]):
                    if verbose: print('B, C = ', B, C)
                    X[i, j] |= set(A for A in V if [B, C] in P[A])
                    if verbose: print('===> X[{}, {}] = {}'.format(i, j, X[i, j]))

    if verbose:
        print('--- end ---')
        for i, j in X:
            print('X[{}, {}] = {}'.format(i, j, X[i, j]))

    return X


def cfg_print_cyk_matrix(X: DefaultDict[Tuple[int, int], Set[Variable]], n: int) -> str:
    def print_set(S: Set[str]) -> str:
        return '{' + ','.join(list(map(str, sorted(S)))) + '}'

    width = (max(len(V) for _, V in X.items()) * 2) + 1
    lines = []
    for i in range(n):
        line = '  '.join("{:<{width}}".format(print_set(X[j-i, j]), width=width) for j in range(i, n))
        # line = '  '.join('X[{},{}]'.format(j-i, j) for j in range(i, n))
        lines.append(line)
    return '\n'.join(reversed(lines))


def cfg_accepts_word(G: CFG, w: str, verbose: bool = False) -> bool:
    """Uses the CYK algorithm to determine if w is in L(G)"""

    if not G.is_chomsky():
        G = cfg_to_chomsky(G)
        if verbose:
            print('grammar converted to Chomsky:')
            print(G)

    R = G.R
    S = G.S

    if w == '':
        return Rule(S, Alternative([])) in R

    X = cfg_cyk_matrix(G, w, verbose)
    n = len(w)
    return S in X[0, n - 1]


# Returns a derivation of the word w
def cfg_derive_word(G: CFG, w: str, derivation_type: str = 'any') -> DerivationTerm:
    assert G.is_chomsky()
    assert derivation_type in ['any', 'leftmost', 'rightmost']

    def find_rule(rules, Xpm, Xmq):
        for BC in rules:
            B, C = BC
            if B in Xpm and C in Xmq:
                return B, C
        return None

    def print_derivation_tree(root):
        S, p, q, children = root
        for child in children:
            print_derivation_tree(child)

    def extract_derivation(root, leftmost: bool):
        S = root[0]
        element = [S]
        result = [element]
        todo = [root]
        while todo:
            A, p, q, children = todo.pop(0) if leftmost else todo.pop()
            if children:
                value = [child[0] for child in children]
                pos = first_index(element, A) if leftmost else last_index(element, A)
                element = element[:pos] + value + element[pos+1:]
                result.append(element)
                if leftmost:
                    todo = children + todo
                else:
                    todo = todo + children
        return result

    R = G.R
    S = G.S

    X = cfg_cyk_matrix(G, w)
    n = len(w)
    if S not in X[0, n - 1]:
        raise RuntimeError("the word '{}' is not accepted by the grammar".format(w))

    # Build a mappings P1 and P2 for productions of length 1 and 2
    P = defaultdict(lambda: [])
    for rule in R:
        P[rule.variable].append(rule.alternative.symbols)

    root = (S, 0, n, [])
    todo = [ root ]
    while todo:
        A, p, q, children = todo.pop()
        if q - p == 1:
            a = w[p]
            child = (a, p, q, [])
            children.append(child)
        else:
            for m in range(p+1, q):
                Xpm = X[p, m-1]
                Xmq = X[m, q-1]
                BC = find_rule(P[A], Xpm, Xmq)
                if not BC:
                    continue
                B, C = BC
                child1 = (B, p, m, [])
                child2 = (C, m, q, [])
                children.append(child1)
                children.append(child2)
                todo.append(child1)
                todo.append(child2)
                break

    # print_derivation_tree(root)
    return extract_derivation(root, derivation_type in ['any', 'leftmost'])


def cfg_words_up_to_n(G: CFG, n: int) -> Set[str]:
    if not G.is_chomsky():
        G = cfg_to_chomsky(G)
    S = G.S
    R = G.R

    words = set([])

    S_empty = Rule(S, Alternative([]))
    if S_empty in R:
        words.add('')

    R1: DefaultDict[Variable, List[Symbol]] = defaultdict(lambda: [])
    R2: DefaultDict[Variable, List[DerivationTerm]] = defaultdict(lambda: [])
    for rule in R:
        if len(rule.alternative.symbols) == 1:
            R1[rule.variable].append(rule.alternative.symbols[0])
        elif len(rule.alternative.symbols) == 2:
            R2[rule.variable].append(rule.alternative.symbols)

    # pre: x.isupper() and len(x) > 0
    # replaces variables by symbols
    def make_words(x: DerivationTerm) -> Set[str]:
        if len(x) == 1:
            return set(R1[x[0]])
        words = make_words(x[1:])
        return set([a + word for a in R1[x[0]] for word in words])

    # pre: x.isupper()
    # replaces one variable by its right hand side
    def replace(x: DerivationTerm) -> List[DerivationTerm]:
        return [x[:j] + rhs + x[j+1:] for j in range(len(x)) for rhs in R2[x[j]]]

    def remove_duplicates(x: list) -> list:
        x.sort()
        return list(k for k, _ in itertools.groupby(x))

    W: List[DerivationTerm] = [[G.S]]
    words = words | make_words([G.S])

    for i in range(2, n + 1):
        W = remove_duplicates([x for word in W for x in replace(word)])
        words = words.union(*[make_words(word) for word in W])

    return words


def cfg_is_simple(G: CFG) -> bool:
    def is_simple_variable(x: str):
        return len(x) == 1 and x.isupper()

    def is_simple_symbol(x: str):
        return len(x) == 1 and x.islower()

    return all(is_simple_variable(x) for x in G.V) and all(is_simple_symbol(x) for x in G.Sigma)


def cfg_productive_variables(G: CFG) -> Set[Variable]:
    productive = set([])

    def is_productive_symbol(x: Union[Variable, Terminal]) -> bool:
        return isinstance(x, Terminal) or x in productive

    while True:
        changed = False
        for r in G.R:
            if r.variable not in productive and all(is_productive_symbol(x) for x in r.alternative.symbols):
                productive.add(r.variable)
                changed = True
        if not changed:
            break
    return productive


def cfg_remove_inproductive_variables_in_place(G: CFG) -> None:
    from gambatools.list_utility import remove_if

    productive = cfg_productive_variables(G)

    def is_productive_symbol(x: Union[Variable, Terminal]) -> bool:
        return isinstance(x, Terminal) or x in productive

    def is_productive(r: Rule) -> bool:
        return r.variable in productive and all(is_productive_symbol(x) for x in r.alternative.symbols)

    remove_if(G.R, lambda r: not is_productive(r))
    G.V &= productive


def cfg_remove_inproductive_variables(G: CFG) -> CFG:
    G = copy.deepcopy(G)
    cfg_remove_inproductive_variables_in_place(G)
    return G


def cfg_remove_useless_rules_in_place(G: CFG) -> None:
    from gambatools.list_utility import remove_if

    def is_useless(r: Rule) -> bool:
        return r.alternative.symbols == [r.variable]

    remove_if(G.R, lambda r: is_useless(r))


def cfg_remove_useless_rules(G: CFG) -> CFG:
    G = copy.deepcopy(G)
    cfg_remove_useless_rules_in_place(G)
    return G


def cfg_print_simple(G):
    if not cfg_is_simple(G):
        raise RuntimeError('the CFG is not in simple format')

    def print_alternative(a: Alternative) -> str:
        if not a.symbols:
            return 'ε'
        return ''.join(a.symbols)

    rule_map = defaultdict(lambda: [])
    for rule in G.R:
        rule_map[rule.variable].append(rule.alternative)
    rules = ['{} -> {}'.format(X, ' | '.join(list(map(print_alternative, rule_map[X])))) for X in G.ordered_variables()]
    return '\n'.join(rules)



#  (C) Copyright Wieger Wesselink 2020. Distributed under the GPL-3.0-or-later
#  Software License, (See accompanying file LICENSE or copy at
#  https://www.gnu.org/licenses/gpl-3.0.txt)

from gambatools.dfa import DFA, State, Symbol

DOT_AUTOMATON_TEXT = '''digraph
{
  fake [style=invisible]
  fake -> "<INITIALSTATE>" [style=bold]

<STATES>

<TRANSITIONS>
}
'''


def dfa_to_dot(D: DFA) -> str:
    def state(q: State) -> str:
        attributes = []
        if q == D.q0:
            attributes.append('root = true')
        if q in D.F:
            attributes.append('shape = doublecircle')
        return '  "{}" [{}]'.format(q, ', '.join(attributes)) if attributes else '  "{}"'.format(q)

    def transition(q: State, a: Symbol, q1: State) -> str:
        return '  "{}" -> "{}" [label="{}"]'.format(q, q1, a)

    text = DOT_AUTOMATON_TEXT
    text = text.replace('<INITIALSTATE>', D.q0)
    text = text.replace('<STATES>', '\n'.join([state(q) for q in D.Q]))
    text = text.replace('<TRANSITIONS>', '\n'.join([transition(q, a, q1) for (q, a), q1 in D.delta.items()]))
    return text
